@@ -26,6 +26,7 @@
 
 #include "argparse.h"
 #include "austin.h"
+#include "hints.h"
 #include "platform.h"
 
 
@@ -39,7 +40,7 @@ const char SAMPLE_FORMAT_ALTERNATIVE[] = ";%s (%s:%d)";
 // Globals for command line arguments
 parsed_args_t pargs = {
   /* t_sampling_interval */ DEFAULT_SAMPLING_INTERVAL,
-  /* timeout             */ DEFAULT_INIT_RETRY_CNT,
+  /* timeout             */ DEFAULT_INIT_RETRY_CNT * 1000,
   /* attach_pid          */ 0,
   /* exclude_empty       */ 0,
   /* sleepless           */ 0,
@@ -49,6 +50,7 @@ parsed_args_t pargs = {
   /* output_file         */ NULL,
   /* output_filename     */ NULL,
   /* children            */ 0,
+  /* exposure            */ 0,
 };
 
 static int exec_arg = 0;
@@ -58,12 +60,83 @@ static int exec_arg = 0;
 
 // ----------------------------------------------------------------------------
 static int
-strtonum(char * str, long * num) {
+str_to_num(char * str, long * num) {
   char * p_err;
 
   *num = strtol(str, &p_err, 10);
 
-  return  (p_err == str || *p_err != 0) ? 1 : 0;
+  return  (p_err == str || *p_err != '\0') ? 1 : 0;
+}
+
+
+/**
+ * Parse the interval argument.
+ *
+ * This acceps s, ms and us as units. The result is in microseconds.
+ */
+static int
+parse_interval(char * str, long * num) {
+  char * p_err;
+
+  *num = strtol(str, &p_err, 10);
+
+  if (p_err == str)
+    FAIL;
+
+  switch (*p_err) {
+  case '\0':
+    SUCCESS;
+  case 's':
+    if (*(p_err+1) != '\0')
+      FAIL;
+    *num = *num * 1000000;
+    break;
+  case 'm':
+    if (*(p_err+1) != 's' || *(p_err+2) != '\0')
+      FAIL;
+    *num = *num * 1000;
+  case 'u':
+    if (*(p_err+1) != 's' || *(p_err+2) != '\0')
+      FAIL;
+    break;
+  default:
+    FAIL;
+  }
+
+  SUCCESS;
+}
+
+
+/**
+ * Parse the timeout argument.
+ *
+ * This acceps s and ms as units. The result is in milliseconds.
+ */
+static int
+parse_timeout(char * str, long * num) {
+  char * p_err;
+
+  *num = strtol(str, &p_err, 10);
+
+  if (p_err == str)
+    FAIL;
+
+  switch (*p_err) {
+  case '\0':
+    SUCCESS;
+  case 's':
+    if (*(p_err+1) != '\0')
+      FAIL;
+    *num = *num * 1000;
+    break;
+  case 'm':
+    if (*(p_err+1) != 's' || *(p_err+2) != '\0')
+      FAIL;
+  default:
+    FAIL;
+  }
+
+  SUCCESS;
 }
 
 
@@ -97,11 +170,11 @@ typedef struct argp_option {
 static struct argp_option options[] = {
   {
     "interval",     'i', "n_us",        0,
-    "Sampling interval (default is 500us)."
+    "Sampling interval in microseconds (default is 100). Accepted units: s, ms, us."
   },
   {
     "timeout",      't', "n_ms",        0,
-    "Approximate start up wait time. Increase on slow machines (default is 100ms)."
+    "Start up wait time in milliseconds (default is 100). Accepted units: s, ms."
   },
   {
     "alt-format",   'a', NULL,          0,
@@ -135,6 +208,10 @@ static struct argp_option options[] = {
     "children",     'C', NULL,          0,
     "Attach to child processes."
   },
+  {
+    "exposure",     'x', "n_sec",       0,
+    "Sample for n_sec seconds only."
+  },
   #ifndef PL_LINUX
   {
     "help",         '?', NULL
@@ -157,6 +234,7 @@ static int
 parse_opt (int key, char *arg, struct argp_state *state)
 {
   if (state->argc == 1) {
+    state->name = PROGRAM_NAME;  // TODO: Check if there are better ways.
     argp_state_help(state, stdout, ARGP_HELP_USAGE);
     exit(0);
   }
@@ -174,7 +252,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
   switch(key) {
   case 'i':
     if (
-      strtonum(arg, (long *) &(pargs.t_sampling_interval)) == 1 ||
+      fail(parse_interval(arg, (long *) &(pargs.t_sampling_interval))) ||
       pargs.t_sampling_interval > LONG_MAX
     )
       argp_error(state, "the sampling interval must be a positive integer");
@@ -182,10 +260,11 @@ parse_opt (int key, char *arg, struct argp_state *state)
 
   case 't':
     if (
-      strtonum(arg, (long *) &(pargs.timeout)) == 1 ||
-      pargs.timeout > LONG_MAX
+      fail(parse_timeout(arg, (long *) &(pargs.timeout))) ||
+      pargs.timeout > LONG_MAX / 1000
     )
       argp_error(state, "timeout must be a positive integer");
+    pargs.timeout *= 1000;
     break;
 
   case 'a':
@@ -209,15 +288,15 @@ parse_opt (int key, char *arg, struct argp_state *state)
     break;
 
   case 'p':
-    if (strtonum(arg, &l_pid) == 1 || l_pid <= 0)
-      argp_error(state, "invalid PID.");
+    if (str_to_num(arg, &l_pid) == 1 || l_pid <= 0)
+      argp_error(state, "invalid PID");
     pargs.attach_pid = (pid_t) l_pid;
     break;
 
   case 'o':
     pargs.output_file = fopen(arg, "w");
     if (pargs.output_file == NULL) {
-      argp_error(state, "Unable to create the given output file.");
+      argp_error(state, "Unable to create the given output file");
     }
     pargs.output_filename = arg;
     break;
@@ -226,10 +305,18 @@ parse_opt (int key, char *arg, struct argp_state *state)
     pargs.children = 1;
     break;
 
+  case 'x':
+    if (
+      str_to_num(arg, (long *) &(pargs.exposure)) == 1 ||
+      pargs.exposure > LONG_MAX
+    )
+      argp_error(state, "the exposure must be a positive integer");
+    break;
+
   case ARGP_KEY_ARG:
   case ARGP_KEY_END:
     if (pargs.attach_pid != 0 && exec_arg != 0)
-      argp_error(state, "the -p option is incompatible with the command argument.");
+      argp_error(state, "the -p option is incompatible with the command argument");
     break;
 
   default:
@@ -359,6 +446,52 @@ _handle_opts(arg_option * opts, arg_callback cb, int * argi, int argc, char ** a
 }
 
 
+static const char * help_msg = \
+"Usage: austin [OPTION...] command [ARG...]\n"
+"Austin -- A frame stack sampler for Python.\n"
+"\n"
+"  -a, --alt-format           Alternative collapsed stack sample format.\n"
+"  -C, --children             Attach to child processes.\n"
+"  -e, --exclude-empty        Do not output samples of threads with no frame\n"
+"                             stacks.\n"
+"  -f, --full                 Produce the full set of metrics (time +mem -mem).\n"
+"  -i, --interval=n_us        Sampling interval in microseconds (default is\n"
+"                             100). Accepted units: s, ms, us.\n"
+"  -m, --memory               Profile memory usage.\n"
+"  -o, --output=FILE          Specify an output file for the collected samples.\n"
+"  -p, --pid=PID              The the ID of the process to which Austin should\n"
+"                             attach.\n"
+"  -s, --sleepless            Suppress idle samples.\n"
+"  -t, --timeout=n_ms         Start up wait time in milliseconds (default is\n"
+"                             100). Accepted units: s, ms.\n"
+"  -x, --exposure=n_sec       Sample for n_sec seconds only.\n"
+"  -?, --help                 Give this help list\n"
+"      --usage                Give a short usage message\n"
+"  -V, --version              Print program version\n"
+"\n"
+"Mandatory or optional arguments to long options are also mandatory or optional\n"
+"for any corresponding short options.\n"
+"\n"
+"Report bugs to <https://github.com/P403n1x87/austin/issues>.\n";
+
+static const char * usage_msg = \
+"Usage: austin [-aCefms?V] [-i n_us] [-o FILE] [-p PID] [-t n_ms] [-x n_sec]\n"
+"            [--alt-format] [--children] [--exclude-empty] [--full]\n"
+"            [--interval=n_us] [--memory] [--output=FILE] [--pid=PID]\n"
+"            [--sleepless] [--timeout=n_ms] [--exposure=n_sec] [--help]\n"
+"            [--usage] [--version] command [ARG...]\n";
+
+
+static void
+arg_error(const char * message) {
+  fputs(PROGRAM_NAME ": ", stderr);
+  fputs(message, stderr);
+  fputc('\n', stderr);
+  fputs("Try `austin --help' or `austin --usage' for more information.\n", stderr);
+  exit(ARG_INVALID_VALUE);
+}
+
+
 // ----------------------------------------------------------------------------
 // Return 0 if all the arguments have been parsed. If interrupted, returns the
 // number of arguments consumed so far. Otherwise return an error code.
@@ -367,6 +500,11 @@ arg_parse(arg_option * opts, arg_callback cb, int argc, char ** argv) {
   int a      = 1;
   int cb_res = 0;
 
+  if (argc <= 1) {
+    puts(usage_msg);
+    exit(0);
+  }
+
   while (a < argc) {
     if (argv[a][0] == '-') {
       if (argv[a][1] == '-') {
@@ -391,61 +529,27 @@ arg_parse(arg_option * opts, arg_callback cb, int argc, char ** argv) {
 }
 
 
-static const char * help_msg = \
-"Usage: austin [OPTION...] command [ARG...]\n"
-"Austin -- A frame stack sampler for Python.\n"
-"\n"
-"  -a, --alt-format           Alternative collapsed stack sample format.\n"
-"  -C, --children             Attach to child processes.\n"
-"  -e, --exclude-empty        Do not output samples of threads with no frame\n"
-"                             stacks.\n"
-"  -f, --full                 Produce the full set of metrics (time +mem -mem).\n"
-"  -i, --interval=n_us        Sampling interval (default is 500us).\n"
-"  -m, --memory               Profile memory usage.\n"
-"  -o, --output=FILE          Specify an output file for the collected samples.\n"
-"  -p, --pid=PID              The the ID of the process to which Austin should\n"
-"                             attach.\n"
-"  -s, --sleepless            Suppress idle samples.\n"
-"  -t, --timeout=n_ms         Approximate start up wait time. Increase on slow\n"
-"                             machines (default is 100ms).\n"
-"  -?, --help                 Give this help list\n"
-"      --usage                Give a short usage message\n"
-"  -V, --version              Print program version\n"
-"\n"
-"Mandatory or optional arguments to long options are also mandatory or optional\n"
-"for any corresponding short options.\n"
-"\n"
-"Report bugs to <https://github.com/P403n1x87/austin/issues>.\n";
-
-static const char * usage_msg = \
-"Usage: austin [-aCefms?V] [-i n_us] [-o FILE] [-p PID] [-t n_ms] [--alt-format]\n"
-"            [--children] [--exclude-empty] [--full] [--interval=n_us]\n"
-"            [--memory] [--output=FILE] [--pid=PID] [--sleepless]\n"
-"            [--timeout=n_ms] [--help] [--usage] [--version] command [ARG...]\n";
-
-
 // ----------------------------------------------------------------------------
 static int
 cb(const char opt, const char * arg) {
   switch (opt) {
   case 'i':
     if (
-      strtonum((char *) arg, (long *) &(pargs.t_sampling_interval)) == 1 ||
+      fail(parse_interval((char *) arg, (long *) &(pargs.t_sampling_interval))) ||
       pargs.t_sampling_interval > LONG_MAX
     ) {
-      puts(usage_msg);
-      return ARG_INVALID_VALUE;
+      arg_error("the sampling interval must be a positive integer");
     }
     break;
 
   case 't':
     if (
-      strtonum((char *) arg, (long *) &(pargs.timeout)) == 1 ||
-      pargs.timeout > LONG_MAX
+      fail(parse_timeout((char *) arg, (long *) &(pargs.timeout))) ||
+      pargs.timeout > LONG_MAX / 1000
     ) {
-      puts(usage_msg);
-      return ARG_INVALID_VALUE;
+      arg_error("the timeout must be a positive integer");
     }
+    pargs.timeout *= 1000;
     break;
 
   case 'a':
@@ -470,11 +574,10 @@ cb(const char opt, const char * arg) {
 
   case 'p':
     if (
-      strtonum((char *) arg, (long *) &pargs.attach_pid) == 1 ||
+      str_to_num((char *) arg, (long *) &pargs.attach_pid) == 1 ||
       pargs.attach_pid <= 0
     ) {
-      puts(usage_msg);
-      return ARG_INVALID_VALUE;
+      arg_error("invalid PID");
     }
     break;
 
@@ -491,6 +594,15 @@ cb(const char opt, const char * arg) {
     pargs.children = 1;
     break;
 
+  case 'x':
+    if (
+      str_to_num((char *) arg, (long *) &(pargs.exposure)) == 1 ||
+      pargs.exposure > LONG_MAX
+    ) {
+      arg_error("the exposure must be a positive integer");
+    }
+    break;
+
   case '?':
     puts(help_msg);
     exit(0);
