@@ -39,10 +39,13 @@
 #include <unistd.h>
 
 #include "argparse.h"
+#include "bin.h"
 #include "dict.h"
 #include "error.h"
+#include "hints.h"
 #include "logging.h"
 #include "mem.h"
+#include "stats.h"
 #include "version.h"
 
 #include "py_proc.h"
@@ -52,15 +55,19 @@
 // ---- PRIVATE ---------------------------------------------------------------
 
 // ---- Retry Timer ----
-#define INIT_RETRY_SLEEP             100   /* us */
-#define INIT_RETRY_CNT                  (pargs.timeout * 1000 / INIT_RETRY_SLEEP)  /* Retry for 0.1s (default) before giving up. */
+#define INIT_RETRY_SLEEP             100   /* μs */
+#define INIT_TIMER_INTERVAL             (pargs.timeout)  /* Retry for 0.1s (default) before giving up. */
 
-#define TIMER_RESET                     (try_cnt=INIT_RETRY_CNT);
-#define TIMER_START                     while (--try_cnt>=0) { usleep(INIT_RETRY_SLEEP);
-#define TIMER_STOP                      (try_cnt = 0);
+#define TIMER_SET(x)                    {_end_time=gettime()+(x);}
+#define TIMER_RESET                     {_end_time=gettime()+INIT_TIMER_INTERVAL;}
+#define TIMER_START                     while(gettime()<=_end_time){usleep(INIT_RETRY_SLEEP);
+#define TIMER_STOP                      {_end_time=0;}
 #define TIMER_END                       }
 
-static int try_cnt;
+static ctime_t _end_time;
+
+
+#define py_proc__memcpy(self, raddr, size, dest)  copy_memory(PROC_REF, raddr, size, dest)
 
 
 // ----------------------------------------------------------------------------
@@ -101,9 +108,7 @@ static const char * _dynsym_array[DYNSYM_COUNT] = {
   "interp_head"
 };
 
-static long _dynsym_hash_array[DYNSYM_COUNT] = {
-  0
-};
+static long _dynsym_hash_array[DYNSYM_COUNT] = {0};
 
 
 #ifdef DEREF_SYM
@@ -111,8 +116,8 @@ static int
 _py_proc__check_sym(py_proc_t * self, char * name, void * value) {
   for (register int i = 0; i < DYNSYM_COUNT; i++) {
     if (
-      string_hash(name) == _dynsym_hash_array[i] &&
-      strcmp(name, _dynsym_array[i]) == 0
+      string_hash(name) == _dynsym_hash_array[i]
+    &&strcmp(name, _dynsym_array[i]) == 0
     ) {
       *(&(self->tstate_curr_raddr) + i) = value;
       log_d("Symbol %s found @ %p", name, value);
@@ -123,28 +128,67 @@ _py_proc__check_sym(py_proc_t * self, char * name, void * value) {
 }
 #endif
 
+
 // ----------------------------------------------------------------------------
 #ifdef PL_UNIX
 #define _popen  popen
 #define _pclose pclose
 #endif
 
+static int
+_get_version_from_executable(char * binary, int * major, int * minor, int * patch) {
+  FILE * fp;
+  char   version[64];
+  char   cmd[256];
+
+  sprintf(cmd, "%s -V 2>&1", binary);
+
+  fp = _popen(cmd, "r");
+  if (!isvalid(fp)) {
+    return NOVERSION;
+  }
+
+  while (fgets(version, sizeof(version) - 1, fp) != NULL) {
+    if (sscanf(version, "Python %d.%d.%d", major, minor, patch) == 3)
+      break;
+  }
+
+  _pclose(fp);
+
+  return (*major << 16) | (*minor << 8);
+}
+
+
 static int
 _py_proc__get_version(py_proc_t * self) {
   if (self == NULL || (self->bin_path == NULL && self->lib_path == NULL))
-    return 0;
+    return NOVERSION;
 
   int major = 0, minor = 0, patch = 0;
 
+  // On Linux, the actual executable is sometimes picked as a library. Hence we
+  // try to execute the library first and see if we get a version from it. If
+  // not, we fall back to the actual binary, if any.
+  #if defined PL_UNIX
+  if (
+    isvalid(self->lib_path)
+  &&_get_version_from_executable(self->lib_path, &major, &minor, &patch) != NOVERSION
+  ) goto from_exe;
+  #endif
+
+  if (
+    isvalid(self->bin_path)
+  &&_get_version_from_executable(self->bin_path, &major, &minor, &patch) != NOVERSION
+  ) goto from_exe;
 
   if (self->bin_path == NULL && self->lib_path != NULL) {
 
     #if defined PL_LINUX                                             /* LINUX */
     if (sscanf(
-        strstr(self->lib_path, "libpython"), "libpython%d.%d", &major, &minor
+        strstr(self->lib_path, "python"), "python%d.%d", &major, &minor
     ) != 2) {
-      log_f("Failed to determine Python version from shared object name.");
-      return 0;
+      set_error(ENOVERSION);
+      return NOVERSION;
     }
 
     #elif defined PL_WIN                                               /* WIN */
@@ -157,37 +201,18 @@ _py_proc__get_version(py_proc_t * self) {
     char * ver_needle = strstr(self->lib_path, "/3.");
     if (ver_needle == NULL) ver_needle = strstr(self->lib_path, "/2.");
     if (ver_needle == NULL || sscanf(ver_needle, "/%d.%d", &major, &minor) != 2) {
-      log_f("Failed to determine Python version from shared object path.");
-      return 0;
+      set_error(ENOVERSION);
+      return NOVERSION;
     }
     #endif
 
-    log_i("Python version: %d.%d.? (from shared library)", major, minor);
-
-    return (major << 16) | (minor << 8);
+    log_m("🐍 Python version: %d.%d.? (from shared library)", major, minor);
+    return VERSION;
   }
 
-
-  FILE *fp;
-  char version[64];
-  char cmd[128];
-
-  sprintf(cmd, "%s -V 2>&1", self->bin_path);
-
-  fp = _popen(cmd, "r");
-  if (fp == NULL) {
-    log_f("Cannot determine the version of Python.");
-    return 0;
-  }
-
-  while (fgets(version, sizeof(version) - 1, fp) != NULL) {
-    if (sscanf(version, "Python %d.%d.%d", &major, &minor, &patch) == 3)
-      break;
-  }
-
-  _pclose(fp);
-
-  log_i("Python version: %d.%d.%d", major, minor, patch);
+from_exe:
+  log_m("🐍 Python version: %d.%d.%d", major, minor, patch);
+  return VERSION;
 
   // Scan the rodata section for something that looks like the Python version.
   // There are good chances this is at the very beginning of the section so
@@ -212,8 +237,6 @@ _py_proc__get_version(py_proc_t * self) {
   //     }
   //   }
   // }
-
-  return (major << 16) | (minor << 8) | patch;
 }
 
 
@@ -231,13 +254,13 @@ _py_proc__check_interp_state(py_proc_t * self, void * raddr) {
       "Cannot copy PyThreadState head at %p from PyInterpreterState instance",
       is.tstate_head
     );
-    return 1;
+    FAIL;
   }
 
   log_t("PyThreadState head loaded @ %p", is.tstate_head);
 
   if (V_FIELD(void*, tstate_head, py_thread, o_interp) != raddr)
-    return 1;
+    FAIL;
 
   log_d(
     "Found possible interpreter state @ %p (offset %p).",
@@ -250,31 +273,21 @@ _py_proc__check_interp_state(py_proc_t * self, void * raddr) {
   );
 
   // As an extra sanity check, verify that the thread state is valid
-  error = EOK;
-  #if defined PL_WIN                                                   /* WIN */
-  raddr_t thread_raddr = { .pid = self->extra->h_proc, .addr = is.tstate_head };
-  #else                                                               /* UNIX */
-  raddr_t thread_raddr = { .pid = self->pid, .addr = is.tstate_head };
-  #endif
-  py_thread_t * thread = py_thread_new_from_raddr(&thread_raddr);
-
-  if (thread == NULL)
-    return 1;
-
-  if (thread->invalid) {
-    py_thread__destroy(thread);
-    log_d("... but Head Thread State is invalid!");
-    return 1;
+  raddr_t thread_raddr = { .pid = PROC_REF, .addr = is.tstate_head };
+  py_thread_t thread;
+  if (fail(py_thread__fill_from_raddr(&thread, &thread_raddr))) {
+    log_d("Failed to fill thread structure");
+    FAIL;
   }
 
-  py_thread__destroy(thread);
+  if (thread.invalid) {
+    log_d("... but Head Thread State is invalid!");
+    FAIL;
+  }
 
-  log_d(
-    "Stack trace constructed from possible interpreter state (error code: %d)",
-    error
-  );
+  log_d("Stack trace constructed from possible interpreter state");
 
-  return error != EOK;
+  SUCCESS;
 }
 
 
@@ -282,8 +295,8 @@ _py_proc__check_interp_state(py_proc_t * self, void * raddr) {
 // ----------------------------------------------------------------------------
 static int
 _py_proc__is_heap_raddr(py_proc_t * self, void * raddr) {
-  if (self == NULL || raddr == NULL || self->map.heap.base == NULL)
-    return 0;
+  if (!isvalid(self) || !isvalid(raddr) || !isvalid(self->map.heap.base))
+    return FALSE;
 
   return (
     raddr >= self->map.heap.base &&
@@ -295,8 +308,8 @@ _py_proc__is_heap_raddr(py_proc_t * self, void * raddr) {
 // ----------------------------------------------------------------------------
 static int
 _py_proc__is_raddr_within_max_range(py_proc_t * self, void * raddr) {
-  if (self == NULL || raddr == NULL || self->map.heap.base == NULL)
-    return 0;
+  if (!isvalid(self) || !isvalid(raddr) || !isvalid(self->map.heap.base))
+    return FALSE;
 
   return (raddr >= self->min_raddr && raddr < self->max_raddr);
 }
@@ -315,14 +328,14 @@ _py_proc__scan_heap(py_proc_t * self) {
     switch (_py_proc__check_interp_state(self, raddr)) {
     case 0:
       self->is_raddr = raddr;
-      return 0;
+      SUCCESS;
 
     case OUT_OF_BOUND:
       return OUT_OF_BOUND;
     }
   }
 
-  return 1;
+  FAIL;
 }
 #endif
 
@@ -330,8 +343,8 @@ _py_proc__scan_heap(py_proc_t * self) {
 // ----------------------------------------------------------------------------
 static int
 _py_proc__scan_bss(py_proc_t * self) {
-  if (py_proc__memcpy(self, self->map.bss.base, self->map.bss.size, self->bss))
-    return 1;
+  if (fail(py_proc__memcpy(self, self->map.bss.base, self->map.bss.size, self->bss)))
+    FAIL;
 
   log_d("Scanning the BSS section for PyInterpreterState");
 
@@ -363,11 +376,11 @@ _py_proc__scan_bss(py_proc_t * self) {
         (void *) raddr - (void *) self->bss
       );
       self->is_raddr = *raddr;
-      return 0;
+      SUCCESS;
     }
   }
 
-  return 1;
+  FAIL;
 }
 
 
@@ -384,7 +397,7 @@ _py_proc__deref_interp_head(py_proc_t * self) {
         "Cannot copy _PyRuntimeState structure from remote address %p",
         self->py_runtime_raddr
       );
-      return 1;
+      FAIL;
     }
     interp_head_raddr = V_FIELD(void *, py_runtime, py_runtime, o_interp_head);
   }
@@ -394,17 +407,17 @@ _py_proc__deref_interp_head(py_proc_t * self) {
         "Cannot copy PyInterpreterState structure from remote address %p",
         self->interp_head_raddr
       );
-      return 1;
+      FAIL;
     }
   }
-  else return 1;
+  else FAIL;
 
   if (_py_proc__check_interp_state(self, interp_head_raddr))
-    return 1;
+    FAIL;
 
   self->is_raddr = interp_head_raddr;
 
-  return 0;
+  SUCCESS;
 }
 
 
@@ -421,23 +434,23 @@ _py_proc__find_interpreter_state(py_proc_t * self) {
     tstate_current_raddr = py_proc__get_current_thread_state_raddr(self);
     if (tstate_current_raddr == NULL || tstate_current_raddr == (void *) -1)
       // Idle or unable to dereference
-      return 1;
+      FAIL;
     else {
-      if (py_proc__get_type(self, tstate_current_raddr, tstate_current))
-        return 1;
+      if (fail(py_proc__get_type(self, tstate_current_raddr, tstate_current)))
+        FAIL;
 
-      if (_py_proc__check_interp_state(
+      if (fail(_py_proc__check_interp_state(
         self, V_FIELD(void*, tstate_current, py_thread, o_interp)
-      )) return 1;
+      ))) FAIL;
 
       self->is_raddr = V_FIELD(void*, tstate_current, py_thread, o_interp);
       log_d("Interpreter head de-referenced from current thread state symbol.");
     }
   } else {
-    log_d("Interpreter head reference from symbol dereferenced successfully.");
+    log_d("✨ Interpreter head de-referenced from symbols ✨ ");
   }
 
-  return 0;
+  SUCCESS;
 
   // 3.6.5 -> 3.6.6: _PyThreadState_Current doesn't seem what one would expect
   //                 anymore, but _PyThreadState_Current.prev is.
@@ -449,23 +462,37 @@ _py_proc__find_interpreter_state(py_proc_t * self) {
       return 1;
     } */
 }
-#endif
+#endif  // DEREF_SYM
 
 
 // ----------------------------------------------------------------------------
 static int
 _py_proc__wait_for_interp_state(py_proc_t * self) {
+  #ifdef DEBUG
   register int attempts = 0;
+  #endif
+
+  self->is_raddr = NULL;
+
   TIMER_RESET
   TIMER_START
+    if (!py_proc__is_running(self)) {
+      set_error(EPROCNPID);
+      FAIL;
+    }
+
+    #ifdef DEBUG
+    attempts++;
+    #endif
+
     #ifdef DEREF_SYM
-    if (_py_proc__find_interpreter_state(self)) {
+    if (fail(_py_proc__find_interpreter_state(self))) {
     #endif
       if (self->bss == NULL) {
         self->bss = malloc(self->map.bss.size);
       }
       if (self->bss == NULL)
-        return 1;
+        FAIL;
 
       switch (_py_proc__scan_bss(self)) {
       case 0:
@@ -479,7 +506,6 @@ _py_proc__wait_for_interp_state(py_proc_t * self) {
       TIMER_STOP
     }
     #endif
-    attempts++;
   TIMER_END
 
   if (self->bss != NULL) {
@@ -499,18 +525,18 @@ _py_proc__wait_for_interp_state(py_proc_t * self) {
       self->is_raddr,
       attempts
     );
-    return 0;
+    SUCCESS;
   }
 
   #ifdef CHECK_HEAP
   log_w("BSS scan unsuccessful so we scan the heap directly ...");
 
   // TODO: Consider copying heap over and check for pointers
-  try_cnt = 10;
+  TIMER_SET(10)
   TIMER_START
     switch (_py_proc__scan_heap(self)) {
     case 0:
-      return 0;
+      SUCCESS;
 
     case OUT_OF_BOUND:
       TIMER_STOP
@@ -518,43 +544,49 @@ _py_proc__wait_for_interp_state(py_proc_t * self) {
   TIMER_END
   #endif
 
-  error = EPROCISTIMEOUT;
-  return 1;
-}
+  set_error(EPROCISTIMEOUT);
+  FAIL;
+} // _py_proc__wait_for_interp_state
 
 
 // ----------------------------------------------------------------------------
 static int
-_py_proc__run(py_proc_t * self) {
-  log_d("Start up timeout: %dms", pargs.timeout);
+_py_proc__run(py_proc_t * self, int try_once) {
+  #ifdef DEBUG
+  if (try_once == FALSE)
+    log_d("Start up timeout: %d ms", pargs.timeout / 1000);
+  else
+    log_d("Single attempt to attach to process %d", self->pid);
+  #endif
 
   TIMER_RESET
   TIMER_START
-    if (self->bin_path != NULL) {
-      free(self->bin_path);
-      self->bin_path = NULL;
+    if (!py_proc__is_running(self)) {
+      set_error(EPROCNPID);
+      FAIL;
     }
 
-    if (self->lib_path != NULL) {
-      free(self->lib_path);
-      self->lib_path = NULL;
-    }
+    sfree(self->bin_path);
+    sfree(self->lib_path);
+    self->sym_loaded = 0;
 
-    if (_py_proc__init(self) == 0)
+    if (success(_py_proc__init(self)))
       break;
 
-    if (error == EPROCPERM || error == EPROCNPID)
-      return 1;  // Fatal errors
+    if (is_fatal(error))
+      FAIL;
 
-    log_d(
-      "Process not ready :: bin_path: %p, lib_path: %p, symbols: %d",
-      self->bin_path, self->lib_path, self->sym_loaded
-    );
+    log_d("Process is not ready");
+
+    if (try_once)
+      TIMER_STOP
   TIMER_END
 
   if (self->bin_path == NULL && self->lib_path == NULL) {
-    log_f("Python binary not found. Not Python?");
-    return 1;
+    if (try_once)
+      log_d("Cannot attach to process %d with a single attempt.", self->pid);
+    set_error(EPROC);
+    FAIL;
   }
 
   if (self->map.bss.size == 0 || self->map.bss.base == NULL)
@@ -585,22 +617,19 @@ _py_proc__run(py_proc_t * self) {
 
   // Determine and set version
   if (!self->version) {
-    self->version = _py_proc__get_version(self);
-    if (!self->version) {
-      log_f("Python version is unknown.");
-      return 1;
-    }
+    if (!(self->version = _py_proc__get_version(self)))
+      FAIL;
 
     set_version(self->version);
   }
 
-  if (_py_proc__wait_for_interp_state(self)) {
-    log_error();
-    return 1;
-  }
+  if (_py_proc__wait_for_interp_state(self))
+    FAIL;
 
-  return 0;
-}
+  self->timestamp = gettime();
+
+  SUCCESS;
+} /* _py_proc__run */
 
 
 // ---- PUBLIC ----------------------------------------------------------------
@@ -609,11 +638,10 @@ _py_proc__run(py_proc_t * self) {
 py_proc_t *
 py_proc_new() {
   py_proc_t * py_proc = (py_proc_t *) calloc(1, sizeof(py_proc_t));
-  if (py_proc == NULL)
-    error = EPROC;
+  if (!isvalid(py_proc))
+    return NULL;
 
-  else
-    py_proc->min_raddr = (void *) -1;
+  py_proc->min_raddr = (void *) -1;
 
   // Pre-hash symbol names
   if (_dynsym_hash_array[0] == 0) {
@@ -623,30 +651,45 @@ py_proc_new() {
   }
 
   py_proc->extra = (proc_extra_info *) calloc(1, sizeof(proc_extra_info));
+  if (!isvalid(py_proc->extra))
+    goto error;
 
-  check_not_null(py_proc);
   return py_proc;
+
+error:
+  free(py_proc);
+  return NULL;
 }
 
 
 // ----------------------------------------------------------------------------
 int
-py_proc__attach(py_proc_t * self, pid_t pid) {
+py_proc__attach(py_proc_t * self, pid_t pid, int child_process) {
   log_d("Attaching to process with PID %d", pid);
 
-  #ifdef PL_WIN                                                        /* WIN */
+  #if defined PL_WIN                                                   /* WIN */
   self->extra->h_proc = OpenProcess(
     PROCESS_VM_READ | PROCESS_QUERY_INFORMATION, FALSE, pid
   );
   if (self->extra->h_proc == INVALID_HANDLE_VALUE) {
-    log_e("Unable to attach to process with PID %d", pid);
-    return 1;
+    set_error(EPROCATTACH);
+    FAIL;
   }
-  #endif
+  #endif                                                               /* ANY */
 
   self->pid = pid;
 
-  return _py_proc__run(self);
+  if (fail(_py_proc__run(self, child_process))) {
+    if (error == EPROCNPID) {
+      set_error(EPROCATTACH);
+    }
+    else {
+      log_ie("Cannot attach to running process.");
+    }
+    FAIL;
+  }
+
+  SUCCESS;
 }
 
 
@@ -668,8 +711,7 @@ py_proc__start(py_proc_t * self, const char * exec, char * argv[]) {
     );
 
     if (nullStdOut == INVALID_HANDLE_VALUE) {
-      log_e("Unable to redirect STDOUT to " NULL_DEVICE);
-      return 1;
+      log_e(error_get_msg(ENULLDEV));
     }
 
     log_d("Redirecting child's STDOUT to " NULL_DEVICE);
@@ -704,8 +746,8 @@ py_proc__start(py_proc_t * self, const char * exec, char * argv[]) {
     free(cmd_line);
 
   if (!process_created) {
-    log_e("Failed to create child process using the command: %s.", exec);
-    return 1;
+    set_error(EPROCFORK);
+    FAIL;
   }
   self->extra->h_proc = piProcInfo.hProcess;
   self->pid = (pid_t) piProcInfo.dwProcessId;
@@ -718,12 +760,11 @@ py_proc__start(py_proc_t * self, const char * exec, char * argv[]) {
     if (pargs.output_file == NULL) {
       log_d("Redirecting child's STDOUT to " NULL_DEVICE);
       if (freopen(NULL_DEVICE, "w", stdout) == NULL)
-        log_e("Unable to redirect child's STDOUT to " NULL_DEVICE);
+        log_e(error_get_msg(ENULLDEV));
     }
 
-    execvp(exec, argv);
+    execvpe(exec, argv, environ);
 
-    log_e("Failed to fork process");
     exit(127);
   }
   #endif                                                               /* ANY */
@@ -737,25 +778,21 @@ py_proc__start(py_proc_t * self, const char * exec, char * argv[]) {
 
   log_d("New process created with PID %d", self->pid);
 
-  return _py_proc__run(self);
-}
+  if (fail(_py_proc__run(self, FALSE))) {
+    if (error == EPROCNPID)
+      set_error(EPROCFORK);
+    log_ie("Cannot start new process");
+    FAIL;
+  }
 
-
-// ----------------------------------------------------------------------------
-int
-py_proc__memcpy(py_proc_t * self, void * raddr, ssize_t size, void * dest) {
-  #if defined PL_WIN                                                   /* WIN */
-  return copy_memory(self->extra->h_proc, raddr, size, dest) == size ? 0 : 1;
-  #else                                                               /* UNIX */
-  return copy_memory(self->pid, raddr, size, dest) == size ? 0 : 1;
-  #endif
+  SUCCESS;
 }
 
 
 // ----------------------------------------------------------------------------
 void
 py_proc__wait(py_proc_t * self) {
-  log_d("Waiting for process to terminate");
+  log_d("Waiting for process %d to terminate", self->pid);
 
   #if defined PL_LINUX
   if (self->extra->wait_thread_id) {
@@ -808,13 +845,13 @@ py_proc__get_current_thread_state_raddr(py_proc_t * self) {
 int
 py_proc__find_current_thread_offset(py_proc_t * self, void * thread_raddr) {
   if (self->py_runtime_raddr == NULL)
-    return 1;
+    FAIL;
 
   void            * interp_head_raddr;
   _PyRuntimeState   py_runtime;
 
   if (py_proc__get_type(self, self->py_runtime_raddr, py_runtime))
-    return 1;
+    FAIL;
 
   interp_head_raddr = V_FIELD(void *, py_runtime, py_runtime, o_interp_head);
 
@@ -837,27 +874,24 @@ py_proc__find_current_thread_offset(py_proc_t * self, void * thread_raddr) {
           "Offset of _PyRuntime.gilstate.tstate_current found at %x",
           self->tstate_current_offset
         );
-        return 0;
+        SUCCESS;
       }
     }
   }
 
-  return 1;
+  FAIL;
 }
 
 
 // ----------------------------------------------------------------------------
 int
 py_proc__is_running(py_proc_t * self) {
-  if (self->is_raddr == NULL)
-    return 0;
-
-  #ifdef PL_WIN                                                        /* WIN */
-  DWORD ec;
-  return GetExitCodeProcess(self->extra->h_proc, &ec) ? ec : -1;
+  #if defined PL_WIN                                                   /* WIN */
+  DWORD ec = 0;
+  return GetExitCodeProcess(self->extra->h_proc, &ec) ? ec == STILL_ACTIVE : 0;
 
   #elif defined PL_MACOS                                             /* MACOS */
-  return pid_to_task(self->pid) != 0;
+  return success(check_pid(self->pid));
 
   #else                                                              /* LINUX */
   return !(kill(self->pid, 0) == -1 && errno == ESRCH);
@@ -865,6 +899,13 @@ py_proc__is_running(py_proc_t * self) {
 }
 
 
+// ----------------------------------------------------------------------------
+int
+py_proc__is_python(py_proc_t * self) {
+  return self->is_raddr != NULL;
+}
+
+
 // ----------------------------------------------------------------------------
 ssize_t
 py_proc__get_memory_delta(py_proc_t * self) {
@@ -877,67 +918,62 @@ py_proc__get_memory_delta(py_proc_t * self) {
 
 
 // ----------------------------------------------------------------------------
-void
+int
 py_proc__sample(py_proc_t * self) {
-  ctime_t   delta;
   ssize_t   mem_delta = 0;
-  void    * current_thread;
-
-  // Compute time delta since last sample.
-  if (!self->timestamp) {
-    self->timestamp = gettime();
-    delta = 0;
-  } else {
-    delta = gettime() - self->timestamp;
-  }
+  void    * current_thread = NULL;
+  ctime_t   delta = gettime() - self->timestamp;  // Time delta since last sample.
 
   PyInterpreterState is;
-  if (py_proc__get_type(self, self->is_raddr, is) != 0)
-    return;
+  if (fail(py_proc__get_type(self, self->is_raddr, is)))
+    FAIL;
 
   if (is.tstate_head != NULL) {
-    #if defined PL_WIN
-    raddr_t raddr = { .pid = self->extra->h_proc, .addr = is.tstate_head };
-    #else
-    raddr_t raddr = { .pid = self->pid, .addr = is.tstate_head };
-    #endif
-    py_thread_t * first_thread = py_thread_new_from_raddr(&raddr);
+    raddr_t raddr = { .pid = PROC_REF, .addr = is.tstate_head };
+    py_thread_t py_thread;
+    if (fail(py_thread__fill_from_raddr(&py_thread, &raddr)))
+      FAIL;
 
     if (pargs.memory) {
       // Use the current thread to determine which thread is manipulating memory
       current_thread = py_proc__get_current_thread_state_raddr(self);
     }
 
-    for (
-      py_thread_t * py_thread = first_thread;
-      py_thread != NULL;
-      py_thread = py_thread__next(py_thread)
-    ) {
+    do {
       if (pargs.memory) {
         mem_delta = 0;
         if (self->py_runtime_raddr != NULL && current_thread == (void *) -1) {
-          if (py_proc__find_current_thread_offset(self, py_thread->raddr.addr))
+          if (py_proc__find_current_thread_offset(self, py_thread.raddr.addr))
             continue;
           else
             current_thread = py_proc__get_current_thread_state_raddr(self);
         }
-        if (py_thread->raddr.addr == current_thread) {
+        if (py_thread.raddr.addr == current_thread) {
           mem_delta = py_proc__get_memory_delta(self);
-          log_t("Thread %lx holds the GIL", py_thread->tid);
+          log_t("Thread %lx holds the GIL", py_thread.tid);
         }
       }
 
-      if (pargs.children)
-        fprintf(pargs.output_file, "Process %d;", self->pid);
-
-      if (!py_thread__print_collapsed_stack(py_thread, delta, mem_delta >> 10))
-        stats_count_sample();
-    }
-
-    py_thread__destroy(first_thread);
+      py_thread__print_collapsed_stack(&py_thread, delta, mem_delta);
+    } while (success(py_thread__next(&py_thread)));
   }
 
   self->timestamp += delta;
+
+  SUCCESS;
+} /* py_proc__sample */
+
+
+// ----------------------------------------------------------------------------
+void
+py_proc__terminate(py_proc_t * self) {
+  if (self->pid) {
+    #if defined PL_UNIX
+    kill(self->pid, SIGTERM);
+    #else
+    TerminateProcess(self->extra->h_proc, 42);
+    #endif
+  }
 }
 
 
